unit Cinema.Net.Protocol.DTO;

{$ifdef FPC}
  {$mode Delphi}
{$endif}

interface

uses
  {$ifdef FPC}
    Classes
  , fpjson
  , jsonparser
  , Cinema.Serialization
  {$else}
    System.Classes
  , System.JSON
  , System.JSON.Serializers
  {$endif}
  ;

type

  {$M+}

  { TCinemaDTO }

  TCinemaDTO = class(TObject)
  private
    class var JsonSerializer: {$ifdef FPC} TJSONStreamer; {$else} TJsonSerializer; {$endif}
    class constructor Create;
    class destructor Destroy;
  private
    FInstanceGuid: AnsiString;
  public
    class function Deserialize<T: class, constructor>(const AJsonDto: WideString): T;
    class function Serialize<T: TCinemaDTO>(ACinemaDto: T; AWrapIntoClassNameField: boolean = false): WideString;
  published
    property InstanceGuid: AnsiString read FInstanceGuid write FInstanceGuid;
  end;

  TCinemaHealthcheckDto = class(TCinemaDTO)
  end;

  TCinemaRemouteInstanceApplicationInfoDto = class(TCinemaDTO)
  private
    FName: UnicodeString;
    FProperties: TJsonObject;
  public
    destructor Destroy; override;
  published
    property Name: UnicodeString read FName write FName;
    property Properties: TJsonObject read FProperties write FProperties;
  end;

  TCinemaBlobDto = class(TCinemaDto)
  private
    FIsLast: boolean;
    FBlob: TMemoryStream;
  public
    destructor Destroy; override;
  published
    property IsLast: boolean read FIsLast;
    property Blob: TMemoryStream read FBlob write FBlob;
  end;

  TCinemaFileDto = class(TCinemaBlobDto)
  private
    FFileName: UnicodeString;
  public
  published
    property FileName: UnicodeString read FFileName write FFileName;
  end;

  TCinemaScenarioDto = class(TCinemaFileDto)
  end;

  TCinemaRunScenarioDto = class(TCinemaDTO)
  end;

  TCinemaMagnetReferenceFileBlockDto = class(TCinemaDTO)
  private
    FHash: Ansistring;
    FPosition: Int64;
    FSize: Int32;
  public
  published
    property Hash: Ansistring read FHash write FHash;
    property Position: Int64 read FPosition write FPosition;
    property Size: Int32 read FSize write FSize;
  end;

  TCinemaMagnetReferenceFileDto = class(TCinemaDTO)
  private
    FBlocks: TArray<TCinemaMagnetReferenceFileBlockDto>;
    FHash: Ansistring;
    FFileName: UnicodeString;
    FSize: UInt64;
  public
    destructor Destroy; override;
    procedure AddFileBlock(AFileBlockDto: TCinemaMagnetReferenceFileBlockDto); overload;
    function AddFileBlock(const AHash: AnsiString; APosition: Int64; ASize: Int32): TCinemaMagnetReferenceFileBlockDto; overload;
  published
    property FileName: UnicodeString read FFileName write FFileName;
    property Hash: Ansistring read FHash write FHash;
    property Blocks: TArray<TCinemaMagnetReferenceFileBlockDto> read FBlocks write FBlocks;
    property Size: UInt64 read FSize write FSize;
  end;

  TCinemaMagnetReferenceDto = class(TCinemaDTO)
  private
    FName: UnicodeString;
    FHash: AnsiString;
    FFiles: TArray<TCinemaMagnetReferenceFileDto>;
    FSubfolder: UnicodeString;
    function GetSize: UInt64;
  public
    destructor Destroy; override;
    procedure AddReferenceFile(AFileDto: TCinemaMagnetReferenceFileDto);
    property Size: UInt64 read GetSize;
  published
    property Name: UnicodeString read FName write FName;
    property Subfolder: UnicodeString read FSubfolder write FSubfolder;
    property Hash: AnsiString read FHash write FHash;
    property Files: TArray<TCinemaMagnetReferenceFileDto> read FFiles write FFiles;
  end;

implementation

{$ifdef FPC}
uses
    Cinema.Utils
  ;
{$else}
uses
    Cinema.Utils
  , System.Rtti
  , System.NetEncoding
  , System.SysUtils
  , System.TypInfo
  , System.JSON.Readers
  , System.JSON.Writers
  ;

type
  TGuidConverter = class(TJsonConverter)
    function CanConvert(ATypeInf: PTypeInfo): Boolean; override;
    function ReadJson(const AReader: TJsonReader; ATypeInf: PTypeInfo; const AExistingValue: TValue;
      const ASerializer: TJsonSerializer): TValue; override;
    procedure WriteJson(const AWriter: TJsonWriter; const AValue: TValue;
      const ASerializer: TJsonSerializer); override;
  end;

  TJsonObjectConverter = class(TJsonConverter)
    function CanConvert(ATypeInf: PTypeInfo): Boolean; override;
    function ReadJson(const AReader: TJsonReader; ATypeInf: PTypeInfo; const AExistingValue: TValue;
      const ASerializer: TJsonSerializer): TValue; override;
    procedure WriteJson(const AWriter: TJsonWriter; const AValue: TValue;
      const ASerializer: TJsonSerializer); override;
  end;

  TBlobConverter = class(TJsonConverter)
    function CanConvert(ATypeInf: PTypeInfo): Boolean; override;
    function ReadJson(const AReader: TJsonReader; ATypeInf: PTypeInfo; const AExistingValue: TValue;
      const ASerializer: TJsonSerializer): TValue; override;
    procedure WriteJson(const AWriter: TJsonWriter; const AValue: TValue;
      const ASerializer: TJsonSerializer); override;
  end;

{ TGuidConverter }

function TGuidConverter.CanConvert(ATypeInf: PTypeInfo): Boolean;
begin
  Result := (ATypeInf.Kind = TTypeKind.tkRecord) and (ATypeInf.Name = 'TGUID');
end;

function TGuidConverter.ReadJson(const AReader: TJsonReader; ATypeInf: PTypeInfo; const AExistingValue: TValue; const ASerializer: TJsonSerializer): TValue;
begin
  Result := TValue.From<TGuid>(StringToGUID(AReader.Value.AsString));
end;

procedure TGuidConverter.WriteJson(const AWriter: TJsonWriter; const AValue: TValue; const ASerializer: TJsonSerializer);
begin
  AWriter.WriteValue(AValue.AsType<TGuid>);
end;

{ TJsonObjectConverter }

function TJsonObjectConverter.CanConvert(ATypeInf: PTypeInfo): Boolean;
begin
  Result := (ATypeInf.Kind = TTypeKind.tkClass) and (ATypeInf.Name = 'TJSONObject');
end;

function TJsonObjectConverter.ReadJson(const AReader: TJsonReader; ATypeInf: PTypeInfo; const AExistingValue: TValue; const ASerializer: TJsonSerializer): TValue;
begin
  Result := TJSONValue.ParseJSONValue(AReader.Value.AsString);
end;

procedure TJsonObjectConverter.WriteJson(const AWriter: TJsonWriter; const AValue: TValue; const ASerializer: TJsonSerializer);
begin
  AWriter.WriteValue(AValue.AsType<TJSONObject>.ToJSON);
end;

{ TBlobConverter }

function TBlobConverter.CanConvert(ATypeInf: PTypeInfo): Boolean;
begin
  Result := (ATypeInf.Kind = TTypeKind.tkClass) and (ATypeInf.Name = 'TMemoryStream');
end;

function TBlobConverter.ReadJson(const AReader: TJsonReader; ATypeInf: PTypeInfo; const AExistingValue: TValue; const ASerializer: TJsonSerializer): TValue;
var
  bytes: TArray<byte>;
  memStr: TMemoryStream;
begin
  bytes := TNetEncoding.Base64String.DecodeStringToBytes(AReader.Value.AsString);
  memStr := TMemoryStream.Create;
  memStr.Write(bytes, length(bytes));
  Result := memStr;
end;

procedure TBlobConverter.WriteJson(const AWriter: TJsonWriter; const AValue: TValue; const ASerializer: TJsonSerializer);
var
  memStr: TMemoryStream;
begin
  inherited;
  memStr := AValue.AsType<TMemoryStream>;
  AWriter.WriteValue(TNetEncoding.Base64String.EncodeBytesToString(memStr.Memory, memStr.Size));
end;
{$endif}

{ TCinemaDTO }

class destructor TCinemaDTO.Destroy;
begin
 JsonSerializer.Destroy;
end;

class function TCinemaDTO.Serialize<T>(ACinemaDto: T; AWrapIntoClassNameField: boolean): WideString;
begin
  Result := JsonSerializer.Serialize<T>(ACinemaDto);
  if AWrapIntoClassNameField then
    Result := '{"' + StringToWide(ClassName) + '":' + Result + '}';
end;

class constructor TCinemaDTO.Create;
begin
  {$ifdef FPC}
  JsonSerializer := TJSONStreamer.Create();
  {$else}
  JsonSerializer := TJsonSerializer.Create;
  JsonSerializer.Converters.Add(TGuidConverter.Create);
  JsonSerializer.Converters.Add(TJsonObjectConverter.Create);
  JsonSerializer.Converters.Add(TBlobConverter.Create);
  {$endif}
end;

class function TCinemaDTO.Deserialize<T>(const AJsonDto: WideString): T;
begin
  Result := JsonSerializer.Deserialize<T>(AJsonDto);
end;

{ TCinemaBlobDto }

destructor TCinemaBlobDto.Destroy;
begin
  FBlob.Free;
  inherited;
end;

{ TCinemaRemouteInstanceApplicationInfoDto }

destructor TCinemaRemouteInstanceApplicationInfoDto.Destroy;
begin
  FProperties.Free;
  inherited;
end;

{ TCinemaMagnetReferenceDto }

procedure TCinemaMagnetReferenceDto.AddReferenceFile(AFileDto: TCinemaMagnetReferenceFileDto);
begin
  SetLength(FFiles, Length(FFiles) + 1);
  FFiles[length(FFiles)-1] := AFileDto;
end;

destructor TCinemaMagnetReferenceDto.Destroy;
var
  f: TCinemaMagnetReferenceFileDto;
begin
  for f in FFiles do
  begin
    f.Free;
  end;
  inherited;
end;

function TCinemaMagnetReferenceDto.GetSize: UInt64;
var
  f: TCinemaMagnetReferenceFileDto;
begin
  Result := 0;
  for f in FFiles do
  begin
    inc(Result, f.Size);
  end;
end;

{ TCinemaMagnetReferenceFileDto }

procedure TCinemaMagnetReferenceFileDto.AddFileBlock(AFileBlockDto: TCinemaMagnetReferenceFileBlockDto);
begin
  SetLength(FBlocks, Length(FBlocks) + 1);
  FBlocks[length(FBlocks)-1] := AFileBlockDto;
end;

function TCinemaMagnetReferenceFileDto.AddFileBlock(const AHash: AnsiString; APosition: Int64; ASize: Int32): TCinemaMagnetReferenceFileBlockDto;
begin
  Result := TCinemaMagnetReferenceFileBlockDto.Create;
  Result.Hash := AHash;
  Result.Position := APosition;
  Result.Size := ASize;
  AddFileBlock(Result);
end;

destructor TCinemaMagnetReferenceFileDto.Destroy;
var
  f: TCinemaMagnetReferenceFileBlockDto;
begin
  for f in FBlocks do
  begin
    f.Free;
  end;
  inherited;
end;

end.
