﻿unit Cinema.Net.SocketCommon;

{$ifdef FPC}
	{$mode delphi}
	{$ifdef posix}
  	{$define fpc_and_posix}
	{$endif}
{$endif}

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
{$IFDEF MSWINDOWS}
  WinSock2,
  {$ifdef FPC}
  Sockets,
  {$endif}
{$ENDIF}

{$IFDEF POSIX}
  {$ifdef FPC}
  baseunix, Unix, Sockets, netdb,
  {$else}
  Posix.NetinetIn, Posix.ArpaInet, Posix.SysSocket, Posix.SysSelect,
  {$endif}
{$ENDIF}
  Classes,
  Types
  ;

const
{$IFDEF MSWINDOWS}
  InvalidSocket = INVALID_SOCKET;
{$ENDIF}
{$IFDEF POSIX}
  InvalidSocket = -1;
{$ENDIF}

type
{$IFDEF MSWINDOWS}
  TSocketHandle = WinSock2.TSocket;
{$ENDIF}
{$IFDEF POSIX}
  TSocketHandle = Integer;
  fd_mask = TSocketHandle;

const
  NFDBITS = Sizeof(fd_mask) shl 3;

{$ENDIF}

type
  PSocketHandle = ^TSocketHandle;

const
  FD_SET_SIZE_DEFAULT = 16;

type

  TCinemaFDSet = class
  private
    FFDSet: PByte;
    FCountSetted: NativeInt;
    FCapacity: Int32;
    FMaxHandle: Int32;
    function GetFdSet: Pointer;
    procedure SetCapacity(const Value: Int32);
    {$IFNDEF POSIX}
    function GetCountEvent: Int32;
    function GetEventHandles(Index: Int32): TSocketHandle;
    {$ENDIF}
  public
    constructor Create(ACapacity: Int32);
    destructor Destroy; override;
    procedure SetFds(ASocketHandles: PSocketHandle; ACount: Int32);
    procedure Clear;
    {$IFDEF POSIX}
    function FdElt(d: Integer): Integer;
    function FdIsSet(fd: Integer): Boolean;
    function FdMask(d: Integer): fd_mask;
    procedure FdSet(fd: Integer);
    {$ENDIF}
    property FdSetPtr: Pointer read GetFdSet;
    property MaxHandle: Int32 read FMaxHandle;
    property Capacity: Int32 read FCapacity write SetCapacity;
    property CountSetted: NativeInt read FCountSetted;
    {$IFNDEF POSIX}
    property CountEvent: Int32 read GetCountEvent;
    property EventHandles[Index: Int32]: TSocketHandle read GetEventHandles;
    {$ENDIF}
  end;

  TSocketType = (TCP, UDP, RAW, RDM, SEQ);

  /// <summary>Wrapper around a network IP address</summary>
  TIPAddress = record
  private
    FAddr: in_addr;
    function GetAddress: string;
  public
    /// <summary>Create and return an ip address structure based on the domain name</summary>
    class function Create(const Name: string): TIPAddress; overload; static;
    class function Create(const Address: Cardinal): TIPAddress; overload; static;
    /// <summary>Create and return an ip address structure based on the platform socket api type</summary>
    class function Create(const Address: in_addr): TIPAddress; overload; static;

    /// <summary>Lookup and return an ip address based on passed in the domain name.
    ///  When lookup failed, then if RaiseException is True then exception will be raised,
    ///  otherwise empty address will be returned. </summary>
    class function LookupName(const Name: string; RaiseException: Boolean = False): TIPAddress; static;
    /// <summary>Return an ip address based on the string representation of an IPv4 address</summary>
    class function LookupAddress(const Address: string): TIPAddress; static;

    class operator Implicit(const Address: TIPAddress): in_addr; inline;

    /// <summary>Address suitable for use with various OS socket apis</summary>
    property Addr: in_addr read FAddr;
    /// <summary>Address in string format suitable for display</summary>
    property Address: string read GetAddress;

  end;

  TNetEndpoint = record
  private
    function GetAddress: TIPAddress;
    function GetFamily: Word;
    function GetPort: Word;
    procedure SetAddress(const Address: TIPAddress); overload;
    procedure SetFamily(const Family: Word);
    procedure SetPort(const Port: Word); overload;
  public
    /// <summary>Create and return a network endpoint based on various address/port forms</summary>
    class function Create(const Name, Service: string; SocketType: TSocketType = TSocketType.TCP): TNetEndpoint; overload; static;
    class function Create(const Address: TIPAddress; const Service: string; SocketType: TSocketType = TSocketType.TCP): TNetEndpoint; overload; static;
    class function Create(const Address: TIPAddress; const Port: Word): TNetEndpoint; overload; static;

    class function Create(const Family: Word; const Name, Service: string; SocketType: TSocketType = TSocketType.TCP): TNetEndpoint; overload; static;
    class function Create(const Family: Word; const Address: TIPAddress; const Service: string; SocketType: TSocketType = TSocketType.TCP): TNetEndpoint; overload; static;
    class function Create(const Family: Word; const Address: TIPAddress; const Port: Word): TNetEndpoint; overload; static;

    class function Create(const Endpoint: sockaddr_in): TNetEndpoint; overload; inline; static;

    class operator Explicit(const EndPoint: TNetEndpoint): sockaddr_in; inline;
    class operator Explicit(const Endpoint: sockaddr_in): TNetEndpoint; inline;
    //class operator Implicit(const Endpoint: sockaddr): TNetEndpoint; inline;

    /// <summary>Lookup a service by name and optionally by the indicated socket type (tcp or udp). Default is tcp.</summary>
    class function LookupService(const AService: string; SocketType: TSocketType = TSocketType.TCP): Word; static;

    /// <summary>Explicitly set the address of the endpoint</summary>
    procedure SetAddress(const Address: string); overload;

    /// <summary>Explicitly set the service port using the name of the service</summary>
    procedure SetPort(const Service: string; SocketType: TSocketType = TSocketType.TCP); overload;

    /// <summary>The IP address contained with this endpoint</summary>
    property Address: TIPAddress read GetAddress write SetAddress;
    /// <summary>Address family of the address contained within this endpoint</summary>
    property Family: Word read GetFamily write SetFamily;
    /// <summary>Port number contained within this endpoint</summary>
    property Port: Word read GetPort write SetPort;

    case byte of
      0: (EndPoint: sockaddr_in);
      1: (Sockaddr: sockaddr);
  end;

  TOnCloseReceiveNotify = procedure(AData: PByte; ALen: Integer) of object;

// system socket API
function CreateSocketHandle(ASocketType: TSocketType): TSocketHandle; overload;
function CreateSocketHandle(ASocketType: TSocketType; ARawFilter: int32): TSocketHandle; overload;
function SetSocketOption(socket: TSocketHandle; level, option_name: Integer; const option_value; option_len: Integer): Integer;
function AcceptSocket(Socket: TSocketHandle; var Addr: sockaddr_in; var Len: Integer): Integer;
//function AcceptSocketBySelect(Socket: TSocketHandle; var Addr: TNetEndpoint; Timeout: UInt32 = 100000): TSocketHandle;
procedure SetReceiveTimeout(ASocket: TSocketHandle; Value: Integer);
procedure SetSendTimeout(ASocket: TSocketHandle; Value: Integer);
function SocketConnect(Socket: TSocketHandle; var Addr: sockaddr; AddrLen: Integer): Integer;
function SocketBind(Socket: TSocketHandle; var Name: sockaddr; NameLen: Integer): Integer;
function SocketListen(Socket: TSocketHandle; QueueSize: Integer): Integer;
function GetRemoteAddress(Socket: TSocketHandle): string;
function GetRemoteEndpoint(Socket: TSocketHandle): TNetEndpoint;
function SocketsSelect(AMaxHandle: TSocketHandle; AFDSetRead, AFDSetWrite, ADSetErrors: Pointer; ATimeOut: uint32): Integer;
function SocketSend(ASocket: TSocketHandle; AData: PByte; ASize: int32): int32;
function SocketSendTo(ASocket: TSocketHandle; const AToAddr: TNetEndpoint; AData: PByte; ASize: int32; AFlags: int32): int32;
function SocketReceive(ASocket: TSocketHandle; ABuffer: PByte; ABufSize: int32; AFlags: uint32): int32;
function SocketReceiveFrom(ASocket: TSocketHandle; ABuffer: PByte; ABufSize: int32; var AddrClient: TNetEndpoint; ATimeOut: uint32): int32;
function CloseSocketHandle(ASocket: TSocketHandle; ASocketType: TSocketType; ABuffer: PByte; ABufferSize: int32; AOnCloseReceiveNotify: TOnCloseReceiveNotify = nil): Integer;
procedure SetAsBroadcastSocket(ASocket: TSocketHandle);
procedure SetNonBlockingMode(ASocket: TSocketHandle; AValue: boolean);
// добавить сокет в мультикастную группу для возможности приёма мультикастных пакетов
// подробнее различия для Windows и Linux см. https://habr.com/ru/articles/141021/
procedure AddMulticastMembership(ASocket: TSocketHandle; const AAddr: TIPAddress);
procedure SetReuseAddrSocketOption(ASocket: TSocketHandle);
//function SetRecieveAllOption(ASocket: TSocketHandle): Integer;
function GetLocalHostName: string;
procedure SetSystemBufferSize(ASocket: TSocketHandle; ARecieve: Boolean; ASize: Integer);

{$ifdef MSWINDOWS}
// в delphi объявлено не верно
function ioctlsocket(s: TSocket; cmd: uint32; var argp: u_long): Integer; stdcall; external 'ws2_32.dll' name 'ioctlsocket';
{$endif}

implementation

uses
  Math
{$IFDEF POSIX}
	{$ifdef FPC}
  , Cinema.Utils
  {$else}
  , Posix.Unistd
  , Posix.SysTime
  , Posix.NetDB
  , Posix.Errno
  , Posix.Signal
  , Posix.Fcntl
  //, Posix.StrOpts
  {$endif}
{$ENDIF}
	;

const
  Proto: array[TSocketType] of Integer = (IPPROTO_TCP, IPPROTO_UDP, IPPROTO_UDP, IPPROTO_IP, IPPROTO_IP);


{ TCinemaFDSet }

procedure TCinemaFDSet.Clear;
begin
  FMaxHandle := 0;
  FCountSetted := 0;
  {$IFDEF POSIX}
  if FCapacity > 0 then
    FillChar(FFDSet^, FCapacity*SizeOf(TSocketHandle), 0);
  {$ELSE}
  if FCapacity > 0 then
    FillChar(FFDSet^, SizeOf(FCountSetted), 0);
  {$ENDIF}
end;

constructor TCinemaFDSet.Create(ACapacity: Int32);
begin
  Capacity := ACapacity;
end;

destructor TCinemaFDSet.Destroy;
begin
  if Assigned(FFDSet) then
    SysFreeMem(FFDSet);
  inherited;
end;

{$IFNDEF POSIX}
function TCinemaFDSet.GetCountEvent: Int32;
begin
  if Assigned(FFDSet) and (FCountSetted > 0) then
    move(FFDSet^, Result, 4)
  else
    Result := 0;
end;

function TCinemaFDSet.GetEventHandles(Index: Int32): TSocketHandle;
var
  count: Int32;
begin
  if (Index > Capacity) or (Index > FCountSetted) then
    exit;

  move(FFDSet^, count, SizeOf(FCountSetted));
  if count <= Index then
    exit;

  move((FFDSet + SizeOf(FCountSetted) + Index*SizeOf(TSocketHandle))^, Result, SizeOf(TSocketHandle));
end;
{$ENDIF}

function TCinemaFDSet.GetFdSet: Pointer;
begin
  Result := FFDSet;
end;

procedure TCinemaFDSet.SetFds(ASocketHandles: PSocketHandle; ACount: Int32);
{$IFDEF POSIX}
var
  i: Int32;
  h: TSocketHandle;
{$ENDIF}
begin
  inc(FCountSetted, ACount);
  if FCountSetted > FCapacity then
    Capacity := (FCountSetted div FD_SET_SIZE_DEFAULT + 1) * FD_SET_SIZE_DEFAULT;
  {$IFDEF POSIX}
  for i := 0 to ACount - 1 do
  begin
    h := PSocketHandle(PByte(ASocketHandles) + i*SizeOf(TSocketHandle))^;
    if h > FMaxHandle then
      FMaxHandle := h;
    FdSet(h);
  end;
  {$ELSE}
  move(FCountSetted, FFDSet[0], SizeOf(FCountSetted));
  move(ASocketHandles^, FFDSet[SizeOf(FCountSetted) + (FCountSetted - ACount)*SizeOf(TSocketHandle)], ACount*SizeOf(TSocketHandle));
  FMaxHandle := FCountSetted;
  {$ENDIF}
end;

procedure TCinemaFDSet.SetCapacity(const Value: Int32);
begin
  if FCapacity = Value then
    exit;
  // для Linux отличается объём в меньшую сторону, т.к. в него фиксируется побитно, пока оставим как есть
  FCapacity := Value;
  if FCapacity > 0 then
  begin
    if Assigned(FFDSet) then
      FFDSet := SysReallocMem(FFDSet, {$IFNDEF POSIX} SizeOf(FCountSetted) + {$ENDIF} FCapacity*SizeOf(TSocketHandle))
    else
      FFDSet := SysGetMem({$IFNDEF POSIX} SizeOf(FCountSetted) + {$ENDIF} FCapacity*SizeOf(TSocketHandle));
  end else
  begin
    SysFreeMem(FFDSet);
    FFDSet := nil;
  end;
end;

{$IFDEF POSIX}

function TCinemaFDSet.FdMask(d: Integer): fd_mask;
begin
  Result := fd_mask(1) shl (d mod NFDBITS);
end;

function TCinemaFDSet.FdIsSet(fd: Integer): Boolean;
begin
  Result := (PSocketHandle(FFDSet+FdElt(fd)*SizeOf(TSocketHandle))^ and FdMask(fd)) <> 0;
end;

function TCinemaFDSet.FdElt(d: Integer): Integer;
begin
  Result := d div NFDBITS;
end;

procedure TCinemaFDSet.FdSet(fd: Integer);
var
  index: Integer;
begin
  index := FdElt(fd)*SizeOf(TSocketHandle);
  PSocketHandle(FFDSet + index)^ := PSocketHandle(FFDSet + index)^ or FdMask(fd);
end;

{$ENDIF}

procedure SetNonBlockingMode(ASocket: TSocketHandle; AValue: boolean);
var
  optVal: {$ifdef posix} Int32 {$else} UInt32 {$endif};
begin
{$ifdef posix}

  optVal := fcntl(ASocket, F_GETFL, 0);
  if (OptVal < 0) then
    exit;
  optVal := IfThen(AValue, optVal or O_NONBLOCK, optVal and not O_NONBLOCK);
  fcntl(ASocket, F_SETFL, optVal);
{$else}
  optVal := IfThen(AValue, 1, 0);
  ioctlsocket(ASocket, FIONBIO, OptVal);
{$endif}
end;

procedure SetAsBroadcastSocket(ASocket: TSocketHandle);
var
  OptVal: Integer;
begin
  OptVal := 1;
  SetSocketOption(ASocket, SOL_SOCKET, SO_BROADCAST, OptVal, SizeOf(OptVal));
end;

procedure SetReuseAddrSocketOption(ASocket: TSocketHandle);
var
  OptVal: Integer;
begin
  OptVal := 1;
  SetSocketOption(ASocket, SOL_SOCKET, SO_REUSEADDR, OptVal, SizeOf(OptVal));
end;

//function SetRecieveAllOption(ASocket: TSocketHandle): Integer;
//const
//  SIO_RCVALL = $98000001;
//var
//  OptVal: Cardinal;
//begin
//  OptVal := 1;
//  {$ifdef posix}
//  {$R-}
//  Result := ioctl(ASocket, SIO_RCVALL, @OptVal);
//  {$R+}
//  {$else}
//  Result := ioctlsocket(ASocket, SIO_RCVALL, OptVal);
//  {$endif}
//end;

procedure AddMulticastMembership(ASocket: TSocketHandle; const AAddr: TIPAddress);
type
  ip_mreq = record
    imr_multiaddr: IN_ADDR;
    imr_interface: IN_ADDR;
  end;

  ip_mreq_source = record
    imr_multiaddr: IN_ADDR;
    imr_sourceaddr: IN_ADDR;
    imr_interface: IN_ADDR;
  end;

const
  {$ifdef posix}
  IP_ADD_MEMBERSHIP = 35;
  IP_ADD_SOURCE_MEMBERSHIP = 39;
  {$endif}
  {$ifdef MSWINDOWS}
  IP_ADD_MEMBERSHIP = 12;
  IP_ADD_SOURCE_MEMBERSHIP = 15;
  {$endif}
var
  mreq: ip_mreq;
begin

  mreq.imr_multiaddr.S_addr := AAddr.Addr.S_addr;             // htonl
  mreq.imr_interface.S_addr := htonl(INADDR_ANY);            // $C0A80102  INADDR_ANY

  if SetSocketOption(ASocket, IPPROTO_IP, IP_ADD_MEMBERSHIP, mreq, sizeof(mreq)) < 0 then
  begin
    mreq.imr_interface.s_addr := INADDR_ANY;
//    var error := GetLastError;
//    error := 0;
  end;
end;

function CreateSocketHandle(ASocketType: TSocketType): TSocketHandle;
const
  _Type: array[TSocketType] of Integer = (SOCK_STREAM, SOCK_DGRAM, SOCK_RAW, SOCK_RDM, SOCK_SEQPACKET);
begin
	{$ifdef FPC}
  Result := fpsocket(PF_INET, _Type[ASocketType], Proto[ASocketType]);
  {$else}
  Result := socket(AF_INET, _Type[ASocketType], Proto[ASocketType]);
  {$endif}
end;

function CreateSocketHandle(ASocketType: TSocketType; ARawFilter: int32): TSocketHandle;
const
  _Type: array[TSocketType] of Integer = (SOCK_STREAM, SOCK_DGRAM, SOCK_RAW, SOCK_RDM, SOCK_SEQPACKET);
begin
	{$ifdef FPC}
  Result := fpsocket(PF_INET, _Type[ASocketType], Proto[ASocketType]);
  {$else}
  Result := socket(AF_INET, _Type[ASocketType], ARawFilter);
  {$endif}
end;

function SetSocketOption(socket: TSocketHandle; level, option_name: Integer; const option_value; option_len: Integer): Integer;
begin
{$IFDEF MSWINDOWS}
  //Result := setsockopt(socket, level, option_name, MarshaledAString(@option_value), option_len);
  Result := setsockopt(socket, level, option_name, @option_value, option_len);
{$ENDIF}
{$IFDEF POSIX}
  {$ifdef FPC}
  	Result := fpsetsockopt(socket, level, option_name, @option_value, option_len);
  {$else}
    Result := setsockopt(socket, level, option_name, option_value, option_len);
  {$endif}
{$ENDIF}
end;

function AcceptSocket(Socket: TSocketHandle; var Addr: sockaddr_in; var Len: Integer): Integer; inline;
begin
{$IFDEF MSWINDOWS}
  {$ifdef FPC}
  	Result := fpaccept(Socket, @Addr, @Len);
	{$else}
  	Result := accept(Socket, @Addr, @Len);
	{$endif}
{$ENDIF}
{$IFDEF POSIX}
  {$ifdef FPC}
	  Result := fpaccept(Socket, psockaddr(@Addr), PSocklen(@Len));
  {$else}
    Result := accept(Socket, sockaddr(Addr), socklen_t(Len));
  {$endif}
{$ENDIF}
end;

//function AcceptSocketBySelect(Socket: TSocketHandle; var Addr: TNetEndpoint; Timeout: UInt32 = 100000): TSocketHandle;
//{$ifdef FPC}
//const
//  INFINITE = dword(-1);
//{$endif}
//var
//  fdAccept: TCinemaFDSet;
//  len: Integer;
//  time: timeval;
//  count: NativeInt;
//begin
//  if Timeout <> INFINITE then
//  begin
//    {$IFDEF POSIX}
//    fdAccept.FdSet(Socket);
//    {$ELSE}
//    fdAccept.fd_count := 1;
//    fdAccept.fd_array[0] := Socket;
//    {$ENDIF}
//    time.tv_sec := Timeout div 1000000;
//    time.tv_usec := Timeout mod 1000000;
//
//    {$ifdef fpc_and_posix}
//    count := fpSelect(Socket + 1, @fdAccept, nil, nil, @time);
//    {$else}
//    count := select(Socket + 1, @fdAccept, nil, nil, @time);
//    {$endif}
//    if count <= 0 then
//      exit(InvalidSocket);
//  end;
//  len := SizeOf(Addr);
//  Result := acceptsocket(Socket, Addr.EndPoint, Len);
//end;

procedure SetSystemBufferSize(ASocket: TSocketHandle; ARecieve: Boolean; ASize: Integer);
begin
  {$ifdef fpc}
  if ARecieve then
    fpsetsockopt(ASocket, SOL_SOCKET, SO_RCVBUF, @ASize, SizeOf(ASize))
  else
    fpsetsockopt(ASocket, SOL_SOCKET, SO_SNDBUF, @ASize, SizeOf(ASize))
  {$else}
  if ARecieve then
    setsockopt(ASocket, SOL_SOCKET, SO_RCVBUF, @ASize, SizeOf(ASize))
  else
    setsockopt(ASocket, SOL_SOCKET, SO_SNDBUF, @ASize, SizeOf(ASize))
  {$endif}
end;

function CloseSocketHandle(ASocket: TSocketHandle; ASocketType: TSocketType; ABuffer: PByte; ABufferSize: int32; AOnCloseReceiveNotify: TOnCloseReceiveNotify = nil): Integer;
const
{$IFDEF MSWINDOWS}
  BOTH = SD_BOTH;
{$ELSE}
  BOTH = SHUT_RDWR;
{$ENDIF}
var
{$IFDEF POSIX}
  res: int32;
{$ENDIF}
  len: int32;
begin
  if (ASocketType in [TSocketType.TCP, TSocketType.RAW, TSocketType.RDM, TSocketType.SEQ]) then
  begin
{$IFDEF POSIX}
    res :=
{$ENDIF}
    {$ifdef fpc}
    fpshutdown(ASocket, BOTH);
    {$else}
    shutdown(ASocket, BOTH);
    {$endif}
{$IFDEF POSIX}
    if (res = 0) or ({$IFDEF FPC}errno {$ELSE} GetLastError {$ENDIF} <> ENOTCONN) then
{$ENDIF}
    begin
      if Assigned(ABuffer) and (ABufferSize > 0) and Assigned(AOnCloseReceiveNotify) then
      begin
        len := SocketReceive(ASocket, ABuffer, ABufferSize, 0);
        while len > 0 do
        begin
          AOnCloseReceiveNotify(ABuffer, len);
          len := SocketReceive(ASocket, ABuffer, ABufferSize, 0);
        end;
      end;
    end;
  end;
{$IFDEF MSWINDOWS}
  Result := closesocket(ASocket);
{$ENDIF}
{$IFDEF POSIX}
  {$ifdef FPC}
	  Result := CloseSocket(ASocket);
  {$else}
  	Result := Posix.Unistd.__close(ASocket);
  {$endif}
{$ENDIF}
end;

procedure SetReceiveTimeout(ASocket: TSocketHandle; Value: Integer);
{$IFDEF POSIX}
var
  time: timeval;
{$ENDIF}
begin
{$IFDEF MSWINDOWS}
  setsocketoption(ASocket, SOL_SOCKET, SO_RCVTIMEO, Value, SizeOf(Value));
{$ENDIF}
{$IFDEF POSIX}
  time.tv_sec := Value div 1000;
  time.tv_usec := (Value mod 1000) * 1000;
  SetSocketOption(ASocket, SOL_SOCKET, SO_RCVTIMEO, time, SizeOf(time));
{$ENDIF}
end;

procedure SetSendTimeout(ASocket: TSocketHandle; Value: Integer);
{$IFDEF POSIX}
var
  time: timeval;
{$ENDIF}
begin
{$IFDEF MSWINDOWS}
  SetSocketOption(ASocket, SOL_SOCKET, SO_SNDTIMEO, Value, SizeOf(Value));
{$ENDIF}
{$IFDEF POSIX}
  time.tv_sec := Value div 1000;
  time.tv_usec := (Value mod 1000) * 1000;
  SetSocketOption(ASocket, SOL_SOCKET, SO_SNDTIMEO, time, SizeOf(time));
{$ENDIF}
end;

function SocketConnect(Socket: TSocketHandle; var Addr: sockaddr; AddrLen: Integer): Integer;
begin
  {$ifdef FPC}
		Result := fpconnect(Socket, @Addr, tsocklen(AddrLen));
  {$else}
  	Result := connect(Socket, Addr, AddrLen);
  {$endif}
end;

function SocketBind(Socket: TSocketHandle; var Name: sockaddr; NameLen: Integer): Integer;
begin
  {$ifdef FPC}
		Result := fpbind(Socket, @Name, tsocklen(NameLen));
  {$else}
  	Result := bind(Socket, Name, NameLen);
  {$endif}
end;

function SocketListen(Socket: TSocketHandle; QueueSize: Integer): Integer;
var
  qSize: Integer;
begin
  qSize := QueueSize;
  if qSize < 0 then
    qSize := SOMAXCONN;

  {$ifdef FPC}
    Result := fplisten(Socket, qSize);
  {$else}
    Result := listen(Socket, qSize);
  {$endif}
end;

function GetRemoteAddress(Socket: TSocketHandle): string;
var
  sa: sockaddr;
  s: {$IFDEF POSIX}uint32{$ELSE}int32{$ENDIF};
begin
  Result := '';
  s := SizeOf(sa);
  {$ifdef FPC}
  if fpgetpeername(Socket, @sa, @s) = 0 then
    Result := string(NetAddrToStr(sockaddr_in(sa).sin_addr));
  {$else}
  if getpeername(Socket, sa, s) = 0 then
    Result := string(inet_ntoa(sockaddr_in(sa).sin_addr));
  {$endif}
end;

function GetRemoteEndpoint(Socket: TSocketHandle): TNetEndpoint;
var
  s: {$IFDEF POSIX}uint32{$ELSE}int32{$ENDIF};
  sa: sockaddr;
begin
  FillChar(sa, SizeOf(sa), 0);
  s := SizeOf(sa);
  {$ifdef FPC}
  if fpgetpeername(Socket, @sa, @s) <> 0 then
  {$else}
  if getpeername(Socket, sa, s) <> 0 then
  {$endif}
    FillChar(Result, SizeOf(Result), 0)
  else
    Result := TNetEndpoint.Create(sockaddr_in(sa));
end;

function GetLocalHostName: string;
{$ifndef fpc_and_posix}
var
  name: array[0..255] of AnsiChar;
{$endif}
begin
  {$ifndef fpc_and_posix}
    Result := '';
    FillChar(name, SizeOf(name), 0);
    if gethostname(@name[0], SizeOf(name)) = 0 then
  	  Result := string(name);
  {$else}
  	Result := GetHostName;
  {$endif}
end;

function SocketsSelect(AMaxHandle: TSocketHandle; AFDSetRead, AFDSetWrite, ADSetErrors: Pointer; ATimeOut: uint32): Integer;
var
  time: timeval;
begin
  time.tv_sec := ATimeOut div 1000000;
  time.tv_usec := ATimeOut mod 1000000;
  {$ifdef fpc_and_posix}
  Result := fpselect(AMaxHandle, AFDSetRead, AFDSetWrite, ADSetErrors, @time);
  {$else}
  Result := select(AMaxHandle, AFDSetRead, AFDSetWrite, ADSetErrors, @time);
  {$endif}
end;

function SocketSendTo(ASocket: TSocketHandle; const AToAddr: TNetEndpoint; AData: PByte; ASize: int32; AFlags: int32): int32;
begin
{$IFDEF MSWINDOWS}
	{$ifdef FPC}
  Result := sendto(ASocket, AData, ASize, AFlags, @AToAddr.Sockaddr, sizeOf(AToAddr.Sockaddr));
  {$else}
  Result := sendto(ASocket, AData^, ASize, AFlags, @AToAddr.Sockaddr, sizeOf(AToAddr.Sockaddr));
	{$endif}
{$ENDIF}
{$IFDEF POSIX}
  {$ifdef FPC}
  Result := fpsendto(ASocket, AData, ASize, AFlags, @AToAddr.Sockaddr, sizeOf(AToAddr.Sockaddr));
  {$else}
  Result := sendto(ASocket, AData^, ASize, AFlags, AToAddr.Sockaddr, sizeOf(AToAddr.Sockaddr));
	{$endif}
{$ENDIF}
end;


function SocketSend(ASocket: TSocketHandle; AData: PByte; ASize: int32): int32;
var
  pos: int32;
  count: int32;
  remain: int32;
begin
  remain := ASize;
  pos := 0;
  while remain > 0 do
  begin
    {$ifdef FPC}
  		count := fpsend(ASocket, @AData[pos], remain, 0);
    {$else}
    	count := send(ASocket, AData[pos], remain, 0);
  	{$endif}
    if count < 0 then
      exit(count);
    inc(pos, count);
    dec(remain, count);
  end;
  Result := pos;
end;

function SocketReceive(ASocket: TSocketHandle; ABuffer: PByte; ABufSize: int32; AFlags: uint32): int32;
begin
  {$ifdef FPC}
  Result := fprecv(ASocket, ABuffer, ABufSize, AFlags);
  {$else}
  Result := recv(ASocket, ABuffer[0], ABufSize, AFlags);
  {$endif}
end;

function SocketReceiveFrom(ASocket: TSocketHandle; ABuffer: PByte; ABufSize: int32; var AddrClient: TNetEndpoint; ATimeOut: uint32): int32;
var
  {$IFDEF POSIX}
  addrLen: socklen_t;
  {$ELSE}
  addrLen: Integer;
  {$ENDIF}
  time: timeval;
begin
  time.tv_sec := ATimeOut div 1000000;
  time.tv_usec := ATimeOut mod 1000000;
  addrLen := SizeOf(addrClient);
  {$ifdef FPC}
  Result := fprecvfrom(ASocket, ABuffer, ABufSize, 0, @AddrClient.Sockaddr, @addrLen);
  {$else}
  Result := recvfrom(ASocket, ABuffer[0], ABufSize, 0, AddrClient.Sockaddr, addrLen);
  {$endif}
end;

{ TIPAddress }

class function TIPAddress.Create(const Name: string): TIPAddress;
begin
  Result := LookupName(Name, True);
end;

class function TIPAddress.Create(const Address: in_addr): TIPAddress;
begin
  Result.FAddr := Address;
end;

class function TIPAddress.Create(const Address: Cardinal): TIPAddress;
begin
  Result.FAddr.S_addr := htonl(Address);
end;

function TIPAddress.GetAddress: string;
begin
  {$ifdef FPC}
  Result := NetAddrToStr(FAddr);
  {$else}
  Result := string(inet_ntoa(FAddr));
  {$endif}
end;

class operator TIPAddress.Implicit(const Address: TIPAddress): in_addr;
begin
  Result := Address.FAddr;
end;

class function TIPAddress.LookupAddress(const Address: string): TIPAddress;
var
  ansiStr: AnsiString;
begin
  ansiStr := AnsiString(Address);
  {$ifdef FPC}
  Result.FAddr.S_addr := StrToNetAddr(ansiStr).s_addr;
  {$else}
  Result.FAddr.S_addr := inet_addr(@ansiStr[1]);
  {$endif}
end;

class function TIPAddress.LookupName(const Name: string; RaiseException: Boolean): TIPAddress;
var
  AsciiStr: AnsiString;
  {$ifndef FPC}
  res: Integer;
  hint: addrinfo;
  HostInfo: Paddrinfo;
  {$endif}
begin
  FillChar(Result, SizeOf(Result), 0);
  if Name = '' then Exit;

  AsciiStr := AnsiString(Name);

  {$ifdef FPC}
   	Result.FAddr := StrToNetAddr(AsciiStr);
  {$else}
    FillChar(hint, SizeOf(hint), 0);
    hint.ai_family := AF_INET;
    {$ifdef LINUX}
      {$WARNINGS OFF}
      hint.ai_flags := AI_IDN;
      {$WARNINGS ON}
    {$endif}
    hint.ai_socktype := SOCK_STREAM;
    HostInfo := nil;

    res := getaddrinfo(@AsciiStr[1], nil, hint, HostInfo);
    if (res = 0) and (HostInfo <> nil) then
    begin
      Result.FAddr := Psockaddr_in(HostInfo.ai_addr).sin_addr;
      freeaddrinfo(HostInfo^);
    end;
  {$endif}
end;

{ TNetEndpoint }

class function TNetEndpoint.Create(const Address: TIPAddress; const Port: Word): TNetEndpoint;
begin
  FillChar(Result.EndPoint, SizeOf(Result.EndPoint), 0);
{$IFDEF MACOS}
  Result.EndPoint.sin_len := SizeOf(EndPoint.EndPoint);
{$ENDIF}
  Result.EndPoint.sin_family := AF_INET;
  Result.EndPoint.sin_addr := Address;
  Result.EndPoint.sin_port := htons(Port);
end;

class function TNetEndpoint.Create(const Address: TIPAddress; const Service: string; SocketType: TSocketType): TNetEndpoint;
begin
  FillChar(Result.EndPoint, SizeOf(Result.EndPoint), 0);
{$IFDEF MACOS}
  Result.EndPoint.sin_len := SizeOf(EndPoint.EndPoint);
{$ENDIF}
  Result.EndPoint.sin_family := AF_INET;
  Result.EndPoint.sin_addr := Address;
  Result.EndPoint.sin_port := htons(LookupService(Service, SocketType));
end;

class function TNetEndpoint.Create(const Name, Service: string; SocketType: TSocketType): TNetEndpoint;
begin
  FillChar(Result.EndPoint, SizeOf(Result.EndPoint), 0);
{$IFDEF MACOS}
  Result.EndPoint.sin_len := SizeOf(EndPoint.EndPoint);
{$ENDIF}
  Result.EndPoint.sin_family := AF_INET;
  Result.EndPoint.sin_addr := TIPAddress.LookupName(Name, True);
  Result.EndPoint.sin_port := htons(LookupService(Service, SocketType));
end;

class function TNetEndpoint.Create(const Family: Word; const Name, Service: string; SocketType: TSocketType): TNetEndpoint;
begin
  FillChar(Result.EndPoint, SizeOf(Result.EndPoint), 0);
{$IFDEF MACOS}
  Result.EndPoint.sin_len := SizeOf(EndPoint.EndPoint);
{$ENDIF}
  Result.EndPoint.sin_family := Family;
  Result.EndPoint.sin_addr := TIPAddress.LookupName(Name, True);
  Result.EndPoint.sin_port := htons(LookupService(Service, SocketType));
end;

class function TNetEndpoint.Create(const Endpoint: sockaddr_in): TNetEndpoint;
begin
{$IFDEF MACOS}
  Result.EndPoint.sin_len := SizeOf(EndPoint.EndPoint);
{$ENDIF}
  Result.EndPoint := Endpoint;
end;

class function TNetEndpoint.Create(const Family: Word; const Address: TIPAddress; const Port: Word): TNetEndpoint;
begin
  FillChar(Result.EndPoint, SizeOf(Result.EndPoint), 0);
{$IFDEF MACOS}
  Result.EndPoint.sin_len := SizeOf(EndPoint.EndPoint);
{$ENDIF}
  Result.EndPoint.sin_family := Family;
  Result.EndPoint.sin_addr := Address;
  Result.EndPoint.sin_port := htons(Port);
end;

class function TNetEndpoint.Create(const Family: Word; const Address: TIPAddress; const Service: string; SocketType: TSocketType): TNetEndpoint;
begin
  FillChar(Result.EndPoint, SizeOf(Result.EndPoint), 0);
{$IFDEF MACOS}
  Result.EndPoint.sin_len := SizeOf(EndPoint.EndPoint);
{$ENDIF}
  Result.EndPoint.sin_family := Family;
  Result.EndPoint.sin_addr := Address;
  Result.EndPoint.sin_port := htons(LookupService(Service, SocketType));
end;

function TNetEndpoint.GetAddress: TIPAddress;
begin
  Result := TIPAddress.Create(EndPoint.sin_addr);
end;

function TNetEndpoint.GetFamily: Word;
begin
  Result := EndPoint.sin_family;
end;

function TNetEndpoint.GetPort: Word;
begin
  Result := ntohs(EndPoint.sin_port);
end;

class operator TNetEndpoint.Explicit(const EndPoint: TNetEndpoint): sockaddr_in;
begin
  Result := EndPoint.EndPoint;
end;

class operator TNetEndpoint.Explicit(const Endpoint: sockaddr_in): TNetEndpoint;
begin
  Result := TNetEndpoint.Create(Endpoint);
end;

//class operator TNetEndpoint.Implicit(const Endpoint: sockaddr): TNetEndpoint;
//begin
//  Result := TNetEndpoint.Create(sockaddr_in(Endpoint));
//end;
//
class function TNetEndpoint.LookupService(const AService: string; SocketType: TSocketType): Word;
const
  Proto: array[TSocketType] of MarshaledAString = ('tcp', 'udp', nil, nil, nil);
var
  {$ifdef fpc_and_posix}
  e: TServiceEntry;
  {$else}
  srvsAnsi: AnsiString;
  servEnt: PServEnt;
  {$endif}
begin
  {$ifdef fpc_and_posix}
  if GetServiceByName(AService, Proto[SocketType], e) then
  	Result := e.Port
  else
    Result := 0;
  {$else}
  srvsAnsi := AnsiString(AService);
  servEnt := getservbyname(@srvsAnsi[1], Proto[SocketType]);
  if Assigned(ServEnt) then
    Result := ntohs(ServEnt.s_port)
  else
    Result := 0;
  {$endif}
end;

procedure TNetEndpoint.SetAddress(const Address: string);
begin
  SetAddress(TIPAddress.LookupName(Address, True));
end;

procedure TNetEndpoint.SetAddress(const Address: TIPAddress);
begin
  EndPoint.sin_addr := Address;
end;

procedure TNetEndpoint.SetFamily(const Family: Word);
begin
  EndPoint.sin_family := Family;
end;

procedure TNetEndpoint.SetPort(const Port: Word);
begin
  EndPoint.sin_port := htons(Port);
end;

procedure TNetEndpoint.SetPort(const Service: string; SocketType: TSocketType);
begin
  SetPort(LookupService(Service, SocketType));
end;

{$ifdef posix}
procedure SignalBrokenPipeHandler(SigNum: Integer); cdecl;
begin
end;
{$endif}

procedure NetInit;
{$IFDEF MSWINDOWS}
var
  Data: WSAData;
{$ENDIF}
begin
{$IFDEF MSWINDOWS}
  WSAStartup(WINSOCK_VERSION, Data);
{$ENDIF}
{$ifdef posix}
  // Глушим ошибку "Broken pipe" возникающая при неожиданном разрыве соединения для одной из сторон
  {$ifdef fpc}
  fpsignal(SIGPIPE, SignalBrokenPipeHandler);
  {$else}
  Posix.Signal.signal(SIGPIPE, SignalBrokenPipeHandler);
  {$endif}
{$endif}
end;

procedure NetUnInit;
begin
{$IFDEF MSWINDOWS}
  WSACleanup;
{$ENDIF}
end;

initialization
  NetInit;


finalization
  NetUnInit;

end.
