﻿unit Cinema.Net.SocketServer;

{$ifdef FPC}
	{$mode delphi}
	{$warn 4046 off}
{$endif}


interface

uses
  Classes,
  Types,
  Cinema.Net.SocketCommon,
  Cinema.Collections,
  Generics.Collections
  {$ifdef fpc}

  {$else}
  //System.Threading
  {$endif}
  ;

type

  TOnReceiveNotify = procedure(AOwner: Pointer; AData: PByte; ALen: Integer) of object;
  TOnCloseNotify = procedure(AOwner: Pointer) of object;
  TOnAcceptNotify = function (ASocket: TSocketHandle; ARemouteEndPoint: TNetEndpoint): Pointer of object;

  TCinemaSocketHandle = record
    Handle: TSocketHandle;
    Owner: Pointer;
    LastTime: uint64;
    FreeOnClose: boolean;
  end;
  PCinemaSocketHandle = ^TCinemaSocketHandle;

  { TSocketList<T> }

  TSocketList<T> = class(TList<T>)
  private
  	type
    	TArrayOfT = array of T;
  {$ifdef fpc}
  private
    function GetList: TArrayOfT;
  public
  	property List: TArrayOfT read GetList;
  {$endif}
  end;

  TListHandles = class(TSocketList<PCinemaSocketHandle>);

  TCinemaSocketServer = class
  private
    const
      RECEIVEBUF_LENGTH = 65536*2;
  private
    FSocketType: TSocketType;
    FEndPoint: TNetEndpoint;
    FRawFilter: int32;
    FDSetRead: TCinemaFDSet;
    //FDSetWrite: TCinemaFDSet;
    FDSetErrors: TCinemaFDSet;
    FHandles: TListHandles;
    FDictHandles: TDictionary<TSocketHandle, PCinemaSocketHandle>;
    FOnReceive: TOnReceiveNotify;
    FOnClose: TOnCloseNotify;
    FBuffer: array[0..RECEIVEBUF_LENGTH-1] of byte;
    FOnAccept: TOnAcceptNotify;
    FArrayHandles: TSocketList<TSocketHandle>;
    FSegmentCount: Integer;
    FHandle: TSocketHandle;
    FOwnHandle: boolean;
    FReceiveTimeout: UInt32;
    FIsRun: Boolean;
    FTimeout: uint32;
    FLastCheckTimeout: uint64;
    FIsBroadcast: Boolean;
    FClosingSocket: PCinemaSocketHandle;
    FNativeIntTag: NativeInt;
    FObjectTag: TObject;
    procedure OnAcceptNewClient(AHandle: TSocketHandle);
    procedure FillSets;
    function ReceiveClient(AFrom: TSocketHandle): boolean; overload; inline;
    procedure SetHandle(const Value: TSocketHandle);
    procedure Accept;
    procedure Listen;
    procedure CloseSocket;
    procedure Reset;
    procedure CheckTimeout;
    procedure OnCloseRecieve(AData: PByte; ALen: Integer);
  public
    constructor Create(const AAddress: string; APort: Word; ASocketType: TSocketType; ARawFilter: int32 = -1);
    destructor Destroy; override;

    ///<summary> Прослушивание подключений и приём данных всех клиентов подключённых к Handle </summary>
    /// <param name="ATimeOut">Таймаут в микросекундах</param>
    function Receive(ATimeOut: uint32 = 10000): boolean; overload;

    ///<summary> приём данных всех клиентов подключённых к Handle </summary>
    /// <param name="ATimeOut">Таймаут в микросекундах</param>
    function ReceiveFrom(ATimeOut: uint32 = 10000): boolean; overload;

    ///<summary> приём данных всех клиентов добавленных в список </summary>
    /// <param name="ATimeOut">Таймаут в микросекундах</param>
    function ReceiveFromClients(ATimeOut: uint32 = 10000): boolean; overload;

    procedure Start(ABeginListen: Boolean);
    procedure Stop;

    function Add(ASocketClientHandle: TSocketHandle; AOwner: Pointer; AFreeOnClose: boolean): PCinemaSocketHandle;
    procedure Remove(ASocketHandle: TSocketHandle);
    function FindHandle(ASocketClientHandle: TSocketHandle): Pointer;

    property Handle: TSocketHandle read FHandle write SetHandle;
    // Подключенные клиенты
    property Clients: TListHandles read FHandles;
    // Количество полученных пакетов
    property SegmentCount: Integer read FSegmentCount;
    property ReceiveTimeout: UInt32 read FReceiveTimeout write FReceiveTimeout;
    property EndPoint: TNetEndpoint read FEndPoint;
    property IsRun: Boolean read FIsRun;
    property IsBroadcast: Boolean read FIsBroadcast;
    { таймаут для UDP клиентов, мс }
    property Timeout: uint32 read FTimeout write FTimeout;
    property SocketType: TSocketType read FSocketType;
    property NativeIntTag: NativeInt read FNativeIntTag write FNativeIntTag;
    property ObjectTag: TObject read FObjectTag write FObjectTag;
    property OwnHandle: boolean read FOwnHandle;

    property OnReceive: TOnReceiveNotify read FOnReceive write FOnReceive;
    property OnClose: TOnCloseNotify read FOnClose write FOnClose;
    property OnAccept: TOnAcceptNotify read FOnAccept write FOnAccept;
  end;

const
  AF_INET = 2;

implementation

uses
  SysUtils
  ;

{ TSocketList<T> }

{$ifdef fpc}
function TSocketList<T>.GetList: TArrayOfT;
begin
	Result := TArrayOfT(FItems);
end;
{$endif}

{ TCinemaSocketServer }

function TCinemaSocketServer.Add(ASocketClientHandle: TSocketHandle; AOwner: Pointer; AFreeOnClose: boolean): PCinemaSocketHandle;
begin
  new(Result);
  Result.Handle := ASocketClientHandle;
  Result.Owner := AOwner;
  Result.LastTime := TThread.GetTickCount64;
  Result.FreeOnClose := AFreeOnClose;
  FHandles.Add(Result);
  FArrayHandles.Add(ASocketClientHandle);
  FDictHandles.Add(ASocketClientHandle, Result);
end;

procedure TCinemaSocketServer.Accept;
const
  Timeout: Cardinal = 1000000;
var
  len: Integer;
  addr: TNetEndpoint;
  client: TSocketHandle;
begin
  len := SizeOf(addr.EndPoint);
  client := AcceptSocket(FHandle, addr.EndPoint, len);
  if client <> InvalidSocket then
  begin
    OnAcceptNewClient(client);
  end;
end;

procedure TCinemaSocketServer.Stop;
begin
  FIsRun := false;
  Reset;
end;

procedure TCinemaSocketServer.Listen;
var
  multicast: TNetEndpoint;
begin

  if FOwnHandle or (FHandle = 0) then
  begin
    FOwnHandle := true;
    if (FRawFilter >= 0) and (FSocketType = TSocketType.RAW) then
      FHandle := CreateSocketHandle(FSocketType, FRawFilter)
    else
      FHandle := CreateSocketHandle(FSocketType);
  end;

  if FHandle = InvalidSocket then
  begin
    //var err := GetLastError;
    FHandle := 0;
    exit;
  end;

  if FIsBroadcast then
  begin
    //SetAsBroadcastSocket(FHandle);
  end;

  SetReuseAddrSocketOption(FHandle);

  if FIsBroadcast then
  begin
    {$ifdef posix}
    multicast := TNetEndpoint.Create(AF_INET, TIPAddress.Create(FEndPoint.Address.Address), FEndPoint.Port);
    {$else}
    multicast := TNetEndpoint.Create(AF_INET, TIPAddress.Create(0), FEndPoint.Port);        //'0.0.0.0'
    {$endif}
    if SocketBind(FHandle, multicast.Sockaddr, SizeOf(multicast)) < 0 then
    begin
      CloseSocket;
      exit;
    end;
    //AddMulticastMembership(FHandle, FEndPoint.Address);
    //SetRecieveAllOption(FHandle);
  end else
  if SocketBind(FHandle, FEndPoint.Sockaddr, SizeOf(FEndPoint)) < 0 then
  begin
    CloseSocket;
    exit;
  end;

  SetNonBlockingMode(FHandle, true);
// пока нет потребности
//  if (FRawFilter >= 0) then
//    SetRecieveAllOption(FHandle);

  if (ReceiveTimeout > 0) then
    SetReceiveTimeout(FHandle, ReceiveTimeout);


  if (FSocketType in [TSocketType.UDP, TSocketType.RAW]) then
    exit;

  if FSocketType in [TSocketType.TCP, TSocketType.RDM, TSocketType.SEQ] then
  begin
    if SocketListen(FHandle, -1) < 0 then
      CloseSocket;
  end;

end;

procedure TCinemaSocketServer.CheckTimeout;
var
  i: int32;
  h: PCinemaSocketHandle;
begin

  FLastCheckTimeout := TThread.GetTickCount64;
  for i := FHandles.Count - 1 downto 0 do
  begin
    h := FHandles.Items[i];
    if FLastCheckTimeout - h.LastTime > FTimeout then
      Remove(h.Handle);
  end;
end;

procedure TCinemaSocketServer.CloseSocket;
begin
  if FOwnHandle then
  begin
    CloseSocketHandle(FHandle, FSocketType, @FBuffer[0], SizeOf(FBuffer));
  end;
  FHandle := 0;
end;

constructor TCinemaSocketServer.Create(const AAddress: string; APort: Word; ASocketType: TSocketType; ARawFilter: int32 = -1);
var
  splittedAddress: TArray<string>;
begin
  FDSetRead := TCinemaFDSet.Create(FD_SET_SIZE_DEFAULT);
  FDSetErrors := TCinemaFDSet.Create(FD_SET_SIZE_DEFAULT);
  FHandles := TListHandles.Create;
  FArrayHandles := TSocketList<TSocketHandle>.Create;
  FDictHandles := TDictionary<TSocketHandle, PCinemaSocketHandle>.Create(8192);
  FEndPoint := TNetEndpoint.Create(AF_INET, TIPAddress.Create(AAddress), APort);
  FSocketType := ASocketType;
  FOwnHandle := true;
  FRawFilter := ARawFilter;
  splittedAddress := AAddress.Split(['.']);
  if length(splittedAddress) = 4 then
    FIsBroadcast := (splittedAddress[3] = '255');
end;

destructor TCinemaSocketServer.Destroy;
begin
  Stop;
  FHandles.Free;
  FArrayHandles.Free;
  FDictHandles.Free;
  FDSetErrors.Free;
  FDSetRead.Free;
  inherited;
end;

procedure TCinemaSocketServer.FillSets;
begin
  FDSetRead.Clear;

  if FHandle > 0 then
  begin
    FDSetRead.SetFds(@FHandle, 1);
  end;

  if FArrayHandles.Count > 0 then
    FDSetRead.SetFds(@FArrayHandles.List[0], FArrayHandles.Count);

  FDSetErrors.Capacity := FDSetRead.CountSetted;
  FDSetErrors.Clear;

end;

function TCinemaSocketServer.FindHandle(ASocketClientHandle: TSocketHandle): Pointer;
var
  h: PCinemaSocketHandle;
begin
  if FDictHandles.TryGetValue(ASocketClientHandle, h) then
    Result := h.Owner
  else
    Result := nil;
end;

procedure TCinemaSocketServer.OnAcceptNewClient(AHandle: TSocketHandle);
var
  clientHolder: Pointer;
begin
  if AHandle > 0 then
  begin
    clientHolder := nil;
    if Assigned(FOnAccept) then
      clientHolder := FOnAccept(AHandle, GetRemoteEndpoint(AHandle));

    if not Assigned(Add(AHandle, clientHolder, true)) then
    begin
      OnClose(clientHolder);
      CloseSocketHandle(AHandle, FSocketType, @FBuffer[0], SizeOf(FBuffer));
    end;
  end;
end;

procedure TCinemaSocketServer.OnCloseRecieve(AData: PByte; ALen: Integer);
begin
  // на случай если кто то решит обработать как строку принятые данные, укажем её конец
  AData[ALen] := 0;
  AData[ALen+1] := 0;
  FOnReceive(FClosingSocket.owner, @AData[0], ALen);
end;

function TCinemaSocketServer.Receive(ATimeOut: uint32): boolean;
begin

  if FHandle = 0 then
  begin
    Listen;
    if FHandle = 0 then
      exit(false);
  end;

  if (FSocketType = TSocketType.UDP) or (FSocketType = TSocketType.RAW) then
  begin
    Result := ReceiveFrom(ATimeOut);
    if (FTimeout > 0) and (TThread.GetTickCount64 - FLastCheckTimeout > 20) then
      CheckTimeout;
    exit;
  end;

  Result := ReceiveFromClients(ATimeOut);
end;

function TCinemaSocketServer.ReceiveFrom(ATimeOut: uint32): boolean;
type
  PNativeUInt = ^NativeUInt;
var
  len: Integer;
	addrClient: TNetEndpoint;
  owner: Pointer;
  client: TSocketHandle;
  h: PCinemaSocketHandle;
begin

  len := SocketReceiveFrom(FHandle, @FBuffer[0], RECEIVEBUF_LENGTH - 2, addrClient, ATimeOut);
  Result := false;
  if ({$ifndef fpc}NativeUInt{$endif}(len) = Integer(InvalidSocket)) or (len = 0) then
  begin
    //Remove(AFrom);
  end else
  begin
    // псевдо-handle которого нам достаточно что бы идентифицировать клиента
    client := PNativeUInt(@addrClient)^;
    if not FDictHandles.TryGetValue(client, h) then
    begin
      if Assigned(FOnAccept) then
        owner := FOnAccept(client, addrClient)
      else
        owner := nil;

      h := Add(client, owner, false);
      if not Assigned(h) then
        exit;
      Result := true;
    end;
    Inc(FSegmentCount);

    h.LastTime := TThread.GetTickCount64;
    // на случай если кто то решит обработать как строку принятые данные, укажем её конец
    FBuffer[len] := 0;
    FBuffer[len+1] := 0;
    FOnReceive(h.owner, @FBuffer[0], len);
    if len > 0 then
      Result := true;
  end;
end;

function TCinemaSocketServer.ReceiveFromClients(ATimeOut: uint32): boolean;
var
  countEvent: Integer;
  i: Integer;
  {$IFNDEF POSIX}
  h: TSocketHandle;
  {$ENDIF}
begin

  Result := false;

  FillSets;
  if FDSetRead.CountSetted = 0 then
    exit;

  countEvent := SocketsSelect(FDSetRead.MaxHandle + 1, FDSetRead.FdSetPtr, nil, FDSetErrors.FdSetPtr, ATimeOut);
  if countEvent < 0 then
  begin
    Reset;
    exit;
  end;

  if countEvent = 0 then
    exit;

  {$IFDEF POSIX}
    if (FHandle > 0) then
    begin
      if FDSetRead.FdIsSet(FHandle) then
      begin
        Result := true;
        Accept;
      end;

      if FDSetErrors.FdIsSet(FHandle) then
      begin
        Reset;
        exit(false);
      end;
    end;

    for i := FHandles.Count - 1 downto 0 do
    begin
      if FDSetRead.FdIsSet(FHandles.Items[i].Handle) then
      begin
        if ReceiveClient(FHandles.Items[i].Handle) then
          Result := true;
      end else
      if FDSetErrors.FdIsSet(FHandles.Items[i].Handle) then
      begin
        Result := true;
        Remove(FHandles.Items[i].Handle);
      end;
    end;

  {$ELSE}

    for i := 0 to FDSetErrors.CountEvent - 1 do
    begin
      Result := true;
      h := FDSetErrors.EventHandles[i];
      if FHandle = h then
      begin
        Reset;
        exit;
      end else
      begin
        Remove(h);
      end;
    end;

    for i := 0 to FDSetRead.CountEvent - 1 do
    begin
      h := FDSetRead.EventHandles[i];
      if h = FHandle then
      begin
        Result := true;
        Accept;
      end else
      if ReceiveClient(h) then
        Result := true;
    end;

  {$ENDIF}
end;

function TCinemaSocketServer.ReceiveClient(AFrom: TSocketHandle): boolean;
var
  len: Integer;
  h: PCinemaSocketHandle;
begin
  len := SocketReceive(AFrom, @FBuffer[0], RECEIVEBUF_LENGTH - 2, 0);  //

  if (len <= 0) then
  begin
    Result := false;
    Remove(AFrom);
  end else
  if len > 0 then
  begin
    Result := true;
    Inc(FSegmentCount);
    FDictHandles.TryGetValue(AFrom, h);
    // на случай если кто то решит обработать как строку принятые данные, укажем её конец
    FBuffer[len] := 0;
    FBuffer[len+1] := 0;
    FOnReceive(h.Owner, @FBuffer[0], len);
  end else
    Result := false;
end;

procedure TCinemaSocketServer.Remove(ASocketHandle: TSocketHandle);
begin
  if FDictHandles.TryGetValue(ASocketHandle, FClosingSocket) then
  begin
    FHandles.Remove(FClosingSocket);
    FDictHandles.Remove(ASocketHandle);
    FArrayHandles.Remove(FClosingSocket.Handle);
    if FClosingSocket.FreeOnClose then
      CloseSocketHandle(ASocketHandle, FSocketType, @FBuffer[0], SizeOf(FBuffer), OnCloseRecieve);
    if Assigned(FOnClose) then
      FOnClose(FClosingSocket.Owner);
    dispose(FClosingSocket);
    FClosingSocket := nil;
  end;
end;

procedure TCinemaSocketServer.Reset;
begin

  if (FHandle > 0) and FOwnHandle then
    CloseSocketHandle(FHandle, FSocketType, @FBuffer[0], SizeOf(FBuffer));

  FHandle := 0;

  while FHandles.Count > 0 do
    Remove(FHandles.Items[FHandles.Count-1].Handle);

end;

procedure TCinemaSocketServer.SetHandle(const Value: TSocketHandle);
begin
  if FHandle = Value then
    exit;

  if (FHandle > 0) and FOwnHandle then
    CloseSocketHandle(FHandle, FSocketType, @FBuffer[0], SizeOf(FBuffer));

  FOwnHandle := false;

  FHandle := Value;
end;

procedure TCinemaSocketServer.Start(ABeginListen: Boolean);
begin
  if FIsRun then
    exit;
  FIsRun := true;
  if ABeginListen then
  begin
    Listen;
  end;
end;

end.
