﻿unit Cinema.PluginsDictionary;

{$ifdef FPC}
	{$mode delphi}
	{$warn 4046 off}
  {$ifdef posix}
    {$define fpc_and_posix}
  {$endif}
{$endif}

interface

uses
  Generics.Collections,
  Cinema.Interfaces
  ;

type

  TPlugin = class;
  TPluginsDictionary = class;
  TCreatePluginExport = function(AName: PWideChar; AConfiguration: PWideChar; ACinema: ICinema): ICinemaPlugin; stdcall;
  TCreatePluginExport2 = function(AName: PWideChar; AConfiguration: PWideChar; ACinema: ICinema; AOutCinemaPlugin: PPointer): boolean; stdcall;

  { TGroupPlugins }

  TGroupPlugins = class
  private
    FName: UnicodeString;
    FPluginDic: TPluginsDictionary;
    FPlugins: TList<TPlugin>;
  public
    constructor Create(APluginDic: TPluginsDictionary; const AName: UnicodeString);
    destructor Destroy; override;
    procedure AddPlugin(APlugin: TPlugin);
    property Name: UnicodeString read FName write FName;
    property PluginDic: TPluginsDictionary read FPluginDic;
    property Plugins: TList<TPlugin> read FPlugins;
  end;

  { TPluginsLibrary }

  TPluginsLibrary = class
  private
    FName: WideString;
    FPath: WideString;
    FPluginsDic: TPluginsDictionary;
    {$ifdef FPC}
    FHandle: TLibHandle;
    {$else}
    FHandle: THandle;
    {$endif}
    // интерфейсные методы библиотеки с плагинами
    GetPluginsListExport: function(ACount: PInteger): PPluginDescription; stdcall;
    CreatePluginExport: TCreatePluginExport;
    CreatePluginExport2: TCreatePluginExport2;
    function GetLibraryPath: WideString;
    function LoadPluginsLibrary: boolean;
  public
    constructor Create(const AName, APath: WideString; APluginsDic: TPluginsDictionary);
    function CreatePlugin(AName: PWideChar; AConfiguration: PWideChar; const ACinema: ICinema): ICinemaPlugin;
    function GetPluginDescription(const AName: WideString): PPluginDescription;
    property Name: WideString read FName;
    property Path: WideString read FPath;
  end;

  { TPlugin }

  TPlugin = class
  private
    FPluginDescription: PPluginDescription;
    FName: UnicodeString;
    FLibrary: TPluginsLibrary;
    FGroupPlugins: TGroupPlugins;

    function GetLibraryPath: WideString;
    function GetPluginDescription: PPluginDescription;
  public
    constructor Create(const AName: WideString; ALibrary: TPluginsLibrary; AGroup: TGroupPlugins);
    destructor Destroy; override;
    { содаёт экземпляр плагина для работы в реальном времени }
    function CreatePlugin(AConfiguration: PWideChar; const ACinema: ICinema): ICinemaPlugin;
    procedure DestroyPlugin(const ACinemaPlugin: ICinemaPlugin);
    property PluginDescription: PPluginDescription read GetPluginDescription;
    property Name: UnicodeString read FName;
    property GroupPlugins: TGroupPlugins read FGroupPlugins write FGroupPlugins;
    property LibraryPath: WideString read GetLibraryPath;
  end;

  TPluginsDictionary = class
  private
    FDictionary: TDictionary<WideString, TPlugin>;
    FDictionaryLibs: TDictionary<WideString, TPluginsLibrary>;
    FGroups: TList<TGroupPlugins>;
    FFileNameDictionary: WideString;
    FSelected: TPlugin;
    FAdditionalSearchPaths: TList<WideString>;
    procedure SetSelected(const Value: TPlugin);
  public
    constructor Create(const AFileNameDictionary: WideString);
    destructor Destroy; override;
    procedure AfterConstruction; override;
    procedure Load;
    procedure Clear;
    procedure AddPluginsSearchPath(const ASearchPath: WideString);
    function GetPlugin(const AName: WideString): TPlugin;
    property Groups: TList<TGroupPlugins> read FGroups;
    property FileNameDictionary: WideString read FFileNameDictionary;
    property Selected: TPlugin read FSelected write SetSelected;
    property AdditionalSearchPaths: TList<WideString> read FAdditionalSearchPaths;
  end;

resourcestring
  sErrorLoadingPackage = 'Can''t load package %s.'+sLineBreak+'%s';

var
  CinemaPlugins: TPluginsDictionary;

implementation

uses
  {$ifdef MSWINDOWS}
  Windows,
  {$endif}

  {$ifdef FPC}
  fpjson,
  Classes,
  dynlibs,
  {$else}
  System.JSON,
  System.IOUtils,
  {$endif}

  SysUtils,
  Cinema.Types,
  Cinema.Utils,
  Cinema.Configuration,
  Cinema.Interfaces.FuncExport
  ;

{ TPluginsDictionary }

procedure TPluginsDictionary.AddPluginsSearchPath(const ASearchPath: WideString);
begin
  FAdditionalSearchPaths.Add(ASearchPath);
end;

procedure TPluginsDictionary.AfterConstruction;
begin
  inherited;
  Load;
end;

procedure TPluginsDictionary.Clear;
var
  pair: TPair<WideString, TPlugin>;
  pairLib: TPair<WideString, TPluginsLibrary>;
  group: TGroupPlugins;
begin
  for pair in FDictionary do
  begin
    pair.Value.Free;
  end;
  FDictionary.Clear;

  for group in FGroups do
  begin
    group.Free;
  end;
  FGroups.Clear;

  for pairLib in FDictionaryLibs do
  begin
    pairLib.Value.Free;
  end;
  FDictionaryLibs.Clear;
end;

constructor TPluginsDictionary.Create(const AFileNameDictionary: WideString);
begin
  FFileNameDictionary := AFileNameDictionary;
  FDictionary := TDictionary<WideString, TPlugin>.Create;
  FGroups := TList<TGroupPlugins>.Create;
  FAdditionalSearchPaths := TList<WideString>.Create;
  FDictionaryLibs := TDictionary<WideString, TPluginsLibrary>.Create;;
end;

destructor TPluginsDictionary.Destroy;
begin
  Clear;
  FGroups.Free;
  FDictionary.Free;
  FDictionaryLibs.Free;
  FAdditionalSearchPaths.Free;
  inherited;
end;

function TPluginsDictionary.GetPlugin(const AName: WideString): TPlugin;
begin
  FDictionary.TryGetValue(AName, Result);
end;

procedure TPluginsDictionary.Load;
const
  csRelativePathLevel = '..' + PathDelim;
var
  i, j, k: int32;
  json, jsnValue, group, plugins, plugin: TJsonCinema;
  grName: WideString;
  plgn: TPlugin;
  libName: UnicodeString;
  path: UnicodeString;
  grp: TGroupPlugins;
  splittedRelativePaht: TArray<UnicodeString>;
  pluginsLib: TPluginsLibrary;
begin
  if not FileExists(FFileNameDictionary) then
    exit;

  json := Cinema.Utils.GetJSON(WideToString(FFileNameDictionary));
  try

    if not Assigned(json) then
      exit;

    jsnValue := json.GetValue<TJsonCinema>('dictionary');
    if not Assigned(jsnValue) then
      exit;

    for i := 0 to TJSONArray(jsnValue).Count - 1 do
    begin
      group := TJSONArray(jsnValue).Items[i];
      grName := group.GetValue<WideString>('Group');
      grp := TGroupPlugins.Create(Self, UnicodeString(grName));
      FGroups.Add(grp);
      plugins := group.GetValue<TJsonCinema>('Plugins');
      for j := 0 to TJsonArray(plugins).Count - 1 do
      begin
      	plugin := TJsonArray(plugins).Items[j];
        libName := plugin.GetValue<WideString>('library');

        if ExtractFileExt(libName) = '' then
          libName := {$ifdef LINUX} '' + {$endif} libName +
            {$ifdef MSWINDOWS} '.dll'{$endif}
            {$ifdef LINUX} '.so'{$endif};

        splittedRelativePaht := libName.Split(['../']);
        if length(splittedRelativePaht) = 0 then
          splittedRelativePaht := libName.Split(['..\']);

        if length(splittedRelativePaht) > 1 then
        begin // путь относительно файла-словаря плагинов (обычно Plugins.json)
          path := ExtractFilePath(FFileNameDictionary);
          for k := 0 to length(splittedRelativePaht) - 2 do
            path := ExtractFilePath(path.TrimRight([PathDelim]));

          splittedRelativePaht[length(splittedRelativePaht)-1] := splittedRelativePaht[length(splittedRelativePaht)-1].Replace({$ifdef posix}'\', PathDelim{$else}'/', PathDelim{$endif});
          // если текущее ПО находится в подкаталоге TCinemaConfiguration.PlatformConfig - для согласованноси пытаемся взять в таком же подкаталоге
          if (TCinemaConfiguration.PlatformConfig <> '') and FileExists(path + ExtractFilePath(splittedRelativePaht[length(splittedRelativePaht)-1]) + TCinemaConfiguration.PlatformConfig + PathDelim + ExtractFileName(splittedRelativePaht[length(splittedRelativePaht)-1])) then
            path := path + ExtractFilePath(splittedRelativePaht[length(splittedRelativePaht)-1]) + TCinemaConfiguration.PlatformConfig + PathDelim + ExtractFileName(splittedRelativePaht[length(splittedRelativePaht)-1])
          else
            path := path + splittedRelativePaht[length(splittedRelativePaht)-1];


        end else
          path := TCinemaConfiguration.RtlPluginsPath + libName;

        if not FDictionaryLibs.TryGetValue(libName, pluginsLib) then
        begin
          pluginsLib := TPluginsLibrary.Create(libName, path, Self);
          FDictionaryLibs.Add(libName, pluginsLib);
        end;
        plgn := TPlugin.Create(plugin.GetValue<WideString>('name'), pluginsLib, grp);
        grp.AddPlugin(plgn);

        FDictionary.Add(WideString(plgn.Name), plgn);
      end;
    end;

  finally
    json.Free;
  end;

end;

procedure TPluginsDictionary.SetSelected(const Value: TPlugin);
begin
  FSelected := Value;
end;

{ TPlugin }

constructor TPlugin.Create(const AName: WideString; ALibrary: TPluginsLibrary; AGroup: TGroupPlugins);
begin
  FName := UnicodeString(AName);
  FLibrary := ALibrary;
  FGroupPlugins := AGroup;
end;

function TPlugin.CreatePlugin(AConfiguration: PWideChar; const ACinema: ICinema): ICinemaPlugin;
begin
  if not Assigned(FPluginDescription) then
    GetPluginDescription;

  if not Assigned(FPluginDescription) then
    exit(nil);

  Result := FLibrary.CreatePlugin(PWideChar(FName), AConfiguration, ACinema);
end;

destructor TPlugin.Destroy;
begin

  inherited;
end;

procedure TPlugin.DestroyPlugin(const ACinemaPlugin: ICinemaPlugin);
begin
  //DestroyPluginExport(ACinemaPlugin);
end;

function TPlugin.GetLibraryPath: WideString;
begin
  Result := FLibrary.Path;
end;

function TPlugin.GetPluginDescription: PPluginDescription;
begin
  if not Assigned(FPluginDescription) then
    FPluginDescription := FLibrary.GetPluginDescription(FName);

  Result := FPluginDescription;
end;

{ TGroupPlugins }

procedure TGroupPlugins.AddPlugin(APlugin: TPlugin);
begin
  FPlugins.Add(APlugin);
end;

constructor TGroupPlugins.Create(APluginDic: TPluginsDictionary; const AName: UnicodeString);
begin
  FPluginDic := APluginDic;
  FPlugins := TList<TPlugin>.Create;
  FName := AName;
end;

destructor TGroupPlugins.Destroy;
begin
  FPlugins.Free;
  inherited;
end;

{$ifdef fpc}
procedure InitializePackage(Module: HMODULE);
type
  TPackageLoad = procedure;
var
  PackageLoad: TPackageLoad;
begin
  //CheckForDuplicateUnits(Module, AValidatePackage);
  PackageLoad := GetProcAddress(Module, 'Initialize'); //Do not localize
  if Assigned(PackageLoad) then
    PackageLoad
  else
    raise Exception.Create('InitializePackage error'); //EPackageError.CreateFmt(sInvalidPackageFile, [GetModuleName(Module)]);
end;

procedure FinalizePackage(Module: HMODULE);
type
  TPackageUnload = procedure;
var
  PackageUnload: TPackageUnload;
begin
  @PackageUnload := GetProcAddress(Module, 'Finalize'); //Do not localize
  if Assigned(PackageUnload) then
    PackageUnload
  else
    raise Exception.Create('FinalizePackage error'); //EPackageError.CreateRes(@sInvalidPackageHandle);
end;

function LoadPackage(const Name: string): HMODULE;
{$IFDEF POSIX}
var
  DLErrorMsg: string;
  M: TMarshaller;
{$ENDIF POSIX}
begin
  Result := SafeLoadLibrary(Name);
  if Result = 0 then
  begin
{$IFDEF POSIX}
    DLErrorMsg := '';//string(dlerror);
{$ENDIF POSIX}
    raise EPackageError.CreateResFmt(@sErrorLoadingPackage,
      [Name,
       {$IFDEF MSWINDOWS}SysErrorMessage(GetLastError){$ENDIF MSWINDOWS}
       {$IFDEF POSIX}DLErrorMsg{$ENDIF POSIX}]);
  end;
  try
    InitializePackage(Result);
  except
    FreeLibrary(Result);
    raise;
  end;
end;

{$endif}

{ TPluginsLibrary }

constructor TPluginsLibrary.Create(const AName, APath: WideString; APluginsDic: TPluginsDictionary);
begin
  FName := AName;
  FPath := APath;
  FPluginsDic := APluginsDic;
end;

function TPluginsLibrary.CreatePlugin(AName: PWideChar; AConfiguration: PWideChar; const ACinema: ICinema): ICinemaPlugin;
var
  plugin: Pointer;
begin
  if (FHandle < 32) and not LoadPluginsLibrary then
    exit(nil);
  plugin := nil;
  TMonitor.Enter(Self);
  try
    if Assigned(CreatePluginExport2) then
    begin
      CreatePluginExport2(AName, AConfiguration, ACinema, @plugin);
    end else
    if Assigned(CreatePluginExport) then
      plugin := CreatePluginExport(AName, AConfiguration, ACinema);
    Result := ICinemaPlugin(plugin);
  finally
    TMonitor.Exit(Self);
  end;
end;

function TPluginsLibrary.GetPluginDescription(const AName: WideString): PPluginDescription;
var
  i: Integer;
  count: Integer;
  plugins: PPluginDescription;
begin
  Result := nil;
  if (FHandle < 32) and not LoadPluginsLibrary then
    exit;
  count := 0;
  plugins := GetPluginsListExport(@count);

  for i := 0 to count - 1 do
  begin
    if WideString(plugins.Name) = AName then
    begin
      Result := plugins;
      break;
    end;
    inc(plugins);
  end;
end;

function TPluginsLibrary.GetLibraryPath: WideString;
var
  p, fn, fnLow: UnicodeString;
begin
  if not FileExists(FPath) then
  begin
    fn := ExtractFileName(FPath);
    fnLow := LowerCase(fn);
    p := ExtractFilePath(FPath);
    Result := p + fnLow;
    if not FileExists(Result) then
    begin
      p := ExtractFilePath(p.TrimRight([PathDelim]));
      Result := p + fnLow;
      if not FileExists(Result) then
      begin
        Result := p + fn;
        if not FileExists(Result) then
        begin
          for p in FPluginsDic.AdditionalSearchPaths do
          begin
            Result := p + fn;
            if FileExists(p) then
              break;
          end;
        end;
      end;
    end;
  end else
    Result := FPath;
end;

function TPluginsLibrary.LoadPluginsLibrary: boolean;
var
  libPath: UnicodeString;
begin
  if FHandle > 32 then
    exit(true);

  libPath := GetLibraryPath;
  try
    if not FileExists(libPath) then
      exit(false);
    if LowerCase(ExtractFileExt(libPath)) = '.bpl' then
    begin
      SetCurrentDir(ExtractFilePath(libPath));
      FHandle := LoadPackage(WideToString(libPath))
    end else
      FHandle := {$ifdef fpc}dynlibs.{$endif}LoadLibrary({$ifndef fpc}PWideChar{$endif}(libPath));
  except
    exit(false);
  end;

  if FHandle < 32 then
    exit(false);

  GetPluginsListExport := GetProcAddress(FHandle, 'GetPluginsList');
  CreatePluginExport := GetProcAddress(FHandle, 'CreatePlugin');
  CreatePluginExport2 := GetProcAddress(FHandle, 'CreatePlugin2');

  Result := true;
end;


initialization

finalization
  CinemaPlugins.Free;

end.
